// Automatically created from `interface.cc.tmp' by glade-filter.awk

#include <unistd-jigdo.h>
/*
 * DO NOT EDIT THIS FILE - it is generated by Glade.
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif
#ifdef GTK_DISABLE_DEPRECATED
#  undef GTK_DISABLE_DEPRECATED
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>

#include <gdk/gdkkeysyms.h>
#include <gtk/gtk.h>

#include <gui.hh>
#include <interface.hh>
#include <gtk/support.hh>

#define GLADE_HOOKUP_OBJECT(component,widget,name) \
  g_object_set_data_full (G_OBJECT (component), "", \
    gtk_widget_ref (widget), (GDestroyNotify) gtk_widget_unref)

#define GLADE_HOOKUP_OBJECT_NO_REF(component,widget,name) \
  g_object_set_data (G_OBJECT (component), "", widget)

GtkWidget*
GUI::Window::create()
{
  GtkWidget *separatortoolitem1;
  GtkWidget *hbox27;
  GtkWidget *vbox9;
  GtkWidget *alignment11;
  GtkWidget *hbox24;
  GtkWidget *vbox8;
  GtkWidget *alignment12;
  GtkWidget *hbox95;
  GtkWidget *image27;
  GtkWidget *label247;
  GtkWidget *table3;
  GtkWidget *scrolledwindow24;
  GtkWidget *viewport6;
  GtkWidget *label156;
  GtkWidget *hbox53;
  GtkWidget *hbox58;
  GtkWidget *hbox59;
  GtkWidget *label161;
  GtkWidget *hbox96;
  GtkWidget *_2lineshighlabelforsettingminimumheight;
  GtkWidget *vbox42;
  GtkWidget *hbox83;
  GtkWidget *image14;
  GtkWidget *image15;
  GtkWidget *label211;
  GtkWidget *image10;
  GtkWidget *image11;
  GtkWidget *image13;
  GtkWidget *scrolledwindow23;
  GtkWidget *viewport5;
  GtkWidget *table5;
  GtkWidget *label151;
  GtkWidget *label131;
  GtkWidget *label133;
  GtkWidget *label149;
  GtkWidget *vbox34;
  GtkWidget *scrolledwindow20;
  GtkWidget *viewport2;
  GtkWidget *label222;
  GtkWidget *hseparator5;
  GtkWidget *hbuttonbox9;
  GtkWidget *button119;
  GtkWidget *alignment3;
  GtkWidget *hbox85;
  GtkWidget *image16;
  GtkWidget *label223;
  GtkWidget *label213;
  GtkWidget *vbox36;
  GtkWidget *vbox39;
  GtkWidget *label229;
  GtkWidget *label230;
  GtkWidget *hbox88;
  GtkWidget *entry23;
  GtkWidget *alignment6;
  GtkWidget *button127;
  GtkWidget *button128;
  GtkWidget *alignment5;
  GtkWidget *hbox89;
  GtkWidget *image23;
  GtkWidget *label227;
  GtkWidget *hbuttonbox10;
  GtkWidget *button120;
  GtkWidget *button121;
  GtkWidget *alignment4;
  GtkWidget *hbox86;
  GtkWidget *image17;
  GtkWidget *label224;
  GtkWidget *label217;
  GtkWidget *vbox40;
  GtkWidget *vbox41;
  GtkWidget *label239;
  GtkWidget *radiobutton1;
radiobutton1_group = NULL;
  GtkWidget *hbox93;
  GtkWidget *radiobutton2;
  GtkWidget *combo1;
  GtkWidget *hbox94;
  GtkWidget *alignment10;
  GtkWidget *radiobutton3;
  GtkWidget *scrolledwindow21;
  GtkWidget *viewport3;
  GtkWidget *table8;
  GtkWidget *label243;
  GtkWidget *label244;
  GtkWidget *label245;
  GtkWidget *label246;
  GtkWidget *combo2;
  GtkWidget *combo3;
  GtkWidget *hbuttonbox11;
  GtkWidget *button131;
  GtkWidget *button132;
  GtkWidget *alignment9;
  GtkWidget *hbox92;
  GtkWidget *image25;
  GtkWidget *label242;
  GtkWidget *label218;
  GtkWidget *vbox37;
  GtkWidget *hbox97;
  GtkWidget *_2lineshighforsettingminheight;
  GtkWidget *vbox43;
  GtkWidget *hbox98;
  GtkWidget *image28;
  GtkWidget *image29;
  GtkWidget *label251;
  GtkWidget *image30;
  GtkWidget *image31;
  GtkWidget *image32;
  GtkWidget *scrolledwindow22;
  GtkWidget *viewport4;
  GtkWidget *table7;
  GtkWidget *label233;
  GtkWidget *label235;
  GtkWidget *label237;
  GtkWidget *label231;
  GtkWidget *label219;

  tooltips = gtk_tooltips_new ();

  window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title (GTK_WINDOW (window), _("Jigsaw Download"));
  gtk_window_set_default_size (GTK_WINDOW (window), 640, 480);

  windowVbox = gtk_vbox_new (FALSE, 2);
  gtk_widget_show (windowVbox);
  gtk_container_add (GTK_CONTAINER (window), windowVbox);

  toolbarHandle = gtk_handle_box_new ();
  gtk_widget_show (toolbarHandle);
  gtk_box_pack_start (GTK_BOX (windowVbox), toolbarHandle, FALSE, FALSE, 0);

  toolbar = gtk_toolbar_new ();
  gtk_widget_show (toolbar);
  gtk_container_add (GTK_CONTAINER (toolbarHandle), toolbar);
  gtk_toolbar_set_style (GTK_TOOLBAR (toolbar), GTK_TOOLBAR_BOTH);
  tmp_toolbar_icon_size = gtk_toolbar_get_icon_size (GTK_TOOLBAR (toolbar));

  toolbarOpen = (GtkWidget*) gtk_tool_button_new_from_stock ("gtk-new");
  gtk_widget_show (toolbarOpen);
  gtk_container_add (GTK_CONTAINER (toolbar), toolbarOpen);
  gtk_container_set_border_width (GTK_CONTAINER (toolbarOpen), 1);
  gtk_widget_set_sensitive (toolbarOpen, FALSE);
  GTK_WIDGET_SET_FLAGS (toolbarOpen, GTK_CAN_FOCUS);
  gtk_tool_item_set_tooltip (GTK_TOOL_ITEM (toolbarOpen), tooltips, _("Open new .jigdo file or URL"), NULL);

  toolbarReuse = (GtkWidget*) gtk_tool_button_new (NULL, _("Reuse files"));
  gtk_container_add (GTK_CONTAINER (toolbar), toolbarReuse);
  gtk_container_set_border_width (GTK_CONTAINER (toolbarReuse), 1);
  gtk_widget_set_sensitive (toolbarReuse, FALSE);
  GTK_WIDGET_SET_FLAGS (toolbarReuse, GTK_CAN_FOCUS);
  gtk_tool_item_set_tooltip (GTK_TOOL_ITEM (toolbarReuse), tooltips, _("Scan local filesystems for data needed by jigsaw downloads - useful when \"upgrading\" e.g. a CD image to a newer version"), NULL);

  separatortoolitem1 = (GtkWidget*) gtk_separator_tool_item_new ();
  gtk_widget_show (separatortoolitem1);
  gtk_container_add (GTK_CONTAINER (toolbar), separatortoolitem1);

  toolbarSettings = (GtkWidget*) gtk_tool_button_new_from_stock ("gtk-properties");
  gtk_container_add (GTK_CONTAINER (toolbar), toolbarSettings);
  gtk_container_set_border_width (GTK_CONTAINER (toolbarSettings), 1);
  gtk_widget_set_sensitive (toolbarSettings, FALSE);
  GTK_WIDGET_SET_FLAGS (toolbarSettings, GTK_CAN_FOCUS);
  gtk_tool_item_set_tooltip (GTK_TOOL_ITEM (toolbarSettings), tooltips, _("Change preferences"), NULL);

  toolbarExit = (GtkWidget*) gtk_tool_button_new_from_stock ("gtk-quit");
  gtk_widget_show (toolbarExit);
  gtk_container_add (GTK_CONTAINER (toolbar), toolbarExit);
  gtk_container_set_border_width (GTK_CONTAINER (toolbarExit), 1);
  GTK_WIDGET_SET_FLAGS (toolbarExit, GTK_CAN_FOCUS);
  gtk_tool_item_set_tooltip (GTK_TOOL_ITEM (toolbarExit), tooltips, _("Exit from Jigsaw Download"), NULL);

  windowPaned = gtk_vpaned_new ();
  gtk_widget_show (windowPaned);
  gtk_box_pack_start (GTK_BOX (windowVbox), windowPaned, TRUE, TRUE, 0);
  gtk_paned_set_position (GTK_PANED (windowPaned), 240);

  invisibleNotebook = gtk_notebook_new ();
  gtk_widget_show (invisibleNotebook);
  gtk_paned_pack1 (GTK_PANED (windowPaned), invisibleNotebook, FALSE, TRUE);
  GTK_WIDGET_SET_FLAGS (invisibleNotebook, GTK_CAN_DEFAULT);
  gtk_notebook_set_show_border (GTK_NOTEBOOK (invisibleNotebook), FALSE);
  gtk_notebook_set_tab_pos (GTK_NOTEBOOK (invisibleNotebook), GTK_POS_BOTTOM);

  pageOpen = gtk_vbox_new (FALSE, 4);
  gtk_widget_show (pageOpen);
  gtk_container_add (GTK_CONTAINER (invisibleNotebook), pageOpen);
  gtk_container_set_border_width (GTK_CONTAINER (pageOpen), 6);

  hbox27 = gtk_hbox_new (FALSE, 0);
  gtk_widget_show (hbox27);
  gtk_box_pack_start (GTK_BOX (pageOpen), hbox27, TRUE, FALSE, 0);
  gtk_container_set_border_width (GTK_CONTAINER (hbox27), 16);

  aboutJigdoLogo = create_pixmap (window, "jigdo-logo.png");
  gtk_widget_show (aboutJigdoLogo);
  gtk_box_pack_start (GTK_BOX (hbox27), aboutJigdoLogo, FALSE, FALSE, 0);

  aboutBgnd = gtk_event_box_new ();
  gtk_widget_show (aboutBgnd);
  gtk_box_pack_start (GTK_BOX (hbox27), aboutBgnd, TRUE, TRUE, 0);

  vbox9 = gtk_vbox_new (FALSE, 0);
  gtk_widget_show (vbox9);
  gtk_container_add (GTK_CONTAINER (aboutBgnd), vbox9);
  gtk_container_set_border_width (GTK_CONTAINER (vbox9), 4);

  aboutJigdoLabel = gtk_label_new (_("#"));
  gtk_widget_show (aboutJigdoLabel);
  gtk_box_pack_start (GTK_BOX (vbox9), aboutJigdoLabel, TRUE, FALSE, 0);
  GTK_WIDGET_SET_FLAGS (aboutJigdoLabel, GTK_CAN_FOCUS);
  gtk_label_set_justify (GTK_LABEL (aboutJigdoLabel), GTK_JUSTIFY_CENTER);
  gtk_label_set_selectable (GTK_LABEL (aboutJigdoLabel), TRUE);

  aboutJigdoButton = gtk_button_new ();
  gtk_widget_show (aboutJigdoButton);
  gtk_box_pack_end (GTK_BOX (vbox9), aboutJigdoButton, FALSE, FALSE, 0);
  GTK_WIDGET_UNSET_FLAGS (aboutJigdoButton, GTK_CAN_FOCUS);
  gtk_button_set_relief (GTK_BUTTON (aboutJigdoButton), GTK_RELIEF_NONE);

  alignment11 = gtk_alignment_new (0.5, 0.5, 1, 1);
  gtk_widget_show (alignment11);
  gtk_container_add (GTK_CONTAINER (aboutJigdoButton), alignment11);

  aboutJigdoButtonLabel = gtk_label_new_with_mnemonic (_("#"));
  gtk_widget_show (aboutJigdoButtonLabel);
  gtk_container_add (GTK_CONTAINER (alignment11), aboutJigdoButtonLabel);
  gtk_label_set_use_markup (GTK_LABEL (aboutJigdoButtonLabel), TRUE);
  gtk_label_set_justify (GTK_LABEL (aboutJigdoButtonLabel), GTK_JUSTIFY_CENTER);

  hbox24 = gtk_hbox_new (FALSE, 4);
  gtk_widget_show (hbox24);
  gtk_box_pack_end (GTK_BOX (pageOpen), hbox24, FALSE, FALSE, 0);

  vbox8 = gtk_vbox_new (FALSE, 0);
  gtk_widget_show (vbox8);
  gtk_box_pack_end (GTK_BOX (hbox24), vbox8, FALSE, FALSE, 0);

  openButton = gtk_button_new ();
  gtk_widget_show (openButton);
  gtk_box_pack_end (GTK_BOX (vbox8), openButton, FALSE, FALSE, 0);
  GTK_WIDGET_SET_FLAGS (openButton, GTK_CAN_DEFAULT);

  alignment12 = gtk_alignment_new (0.5, 0.5, 0, 0);
  gtk_widget_show (alignment12);
  gtk_container_add (GTK_CONTAINER (openButton), alignment12);

  hbox95 = gtk_hbox_new (FALSE, 2);
  gtk_widget_show (hbox95);
  gtk_container_add (GTK_CONTAINER (alignment12), hbox95);
  gtk_container_set_border_width (GTK_CONTAINER (hbox95), 5);

  image27 = gtk_image_new_from_stock ("gtk-ok", GTK_ICON_SIZE_BUTTON);
  gtk_widget_show (image27);
  gtk_box_pack_start (GTK_BOX (hbox95), image27, FALSE, FALSE, 0);

  label247 = gtk_label_new_with_mnemonic (_("<b>_Start</b>"));
  gtk_widget_show (label247);
  gtk_box_pack_start (GTK_BOX (hbox95), label247, FALSE, FALSE, 0);
  gtk_label_set_use_markup (GTK_LABEL (label247), TRUE);

  table3 = gtk_table_new (2, 3, FALSE);
  gtk_widget_show (table3);
  gtk_box_pack_end (GTK_BOX (hbox24), table3, TRUE, TRUE, 4);
  gtk_table_set_row_spacings (GTK_TABLE (table3), 2);
  gtk_table_set_col_spacings (GTK_TABLE (table3), 4);

  open_sourceURLLabel = gtk_label_new (_("URL"));
  gtk_widget_show (open_sourceURLLabel);
  gtk_table_attach (GTK_TABLE (table3), open_sourceURLLabel, 0, 1, 0, 1,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);
  gtk_label_set_justify (GTK_LABEL (open_sourceURLLabel), GTK_JUSTIFY_RIGHT);
  gtk_misc_set_alignment (GTK_MISC (open_sourceURLLabel), 1, 0.5);
  gtk_misc_set_padding (GTK_MISC (open_sourceURLLabel), 2, 0);

  open_destinationLabel = gtk_label_new (_("Save to"));
  gtk_widget_show (open_destinationLabel);
  gtk_table_attach (GTK_TABLE (table3), open_destinationLabel, 0, 1, 1, 2,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);
  gtk_label_set_justify (GTK_LABEL (open_destinationLabel), GTK_JUSTIFY_RIGHT);
  gtk_misc_set_alignment (GTK_MISC (open_destinationLabel), 1, 0.5);
  gtk_misc_set_padding (GTK_MISC (open_destinationLabel), 2, 0);

  open_URL = gtk_entry_new ();
  gtk_widget_show (open_URL);
  gtk_table_attach (GTK_TABLE (table3), open_URL, 1, 2, 0, 1,
                    (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);
  gtk_tooltips_set_tip (tooltips, open_URL, _("If the string you enter does not point to a .jigdo file, jigdo behaves just like a normal \"download manager\" and saves the file in the destination directory, nothing else."), NULL);

  open_dest = gtk_entry_new ();
  gtk_widget_show (open_dest);
  gtk_table_attach (GTK_TABLE (table3), open_dest, 1, 2, 1, 2,
                    (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);
  gtk_tooltips_set_tip (tooltips, open_dest, _("If you entered a .jigdo URL above, this must be a directory name. For single-file downloads, it can also be a filename."), NULL);

  open_URLSel = gtk_button_new_with_mnemonic (_("Browse..."));
  gtk_widget_show (open_URLSel);
  gtk_table_attach (GTK_TABLE (table3), open_URLSel, 2, 3, 0, 1,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);

  open_destSel = gtk_button_new_with_mnemonic (_("Browse..."));
  gtk_widget_show (open_destSel);
  gtk_table_attach (GTK_TABLE (table3), open_destSel, 2, 3, 1, 2,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);

  enterUrlLabel = gtk_label_new (_("Enter location of <tt>.jigdo</tt> file, and destination directory to save files to."));
  gtk_widget_show (enterUrlLabel);
  gtk_box_pack_end (GTK_BOX (pageOpen), enterUrlLabel, FALSE, FALSE, 0);
  gtk_label_set_use_markup (GTK_LABEL (enterUrlLabel), TRUE);

  pageOpenLabel = gtk_label_new (_("#Open"));
  gtk_widget_show (pageOpenLabel);
  gtk_notebook_set_tab_label (GTK_NOTEBOOK (invisibleNotebook), gtk_notebook_get_nth_page (GTK_NOTEBOOK (invisibleNotebook), 0), pageOpenLabel);
  gtk_label_set_justify (GTK_LABEL (pageOpenLabel), GTK_JUSTIFY_CENTER);

  pageReuse = gtk_vbox_new (FALSE, 4);
  gtk_widget_show (pageReuse);
  gtk_container_add (GTK_CONTAINER (invisibleNotebook), pageReuse);
  gtk_container_set_border_width (GTK_CONTAINER (pageReuse), 6);

  scrolledwindow24 = gtk_scrolled_window_new (NULL, NULL);
  gtk_widget_show (scrolledwindow24);
  gtk_box_pack_start (GTK_BOX (pageReuse), scrolledwindow24, TRUE, TRUE, 0);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolledwindow24), GTK_POLICY_NEVER, GTK_POLICY_ALWAYS);

  viewport6 = gtk_viewport_new (NULL, NULL);
  gtk_widget_show (viewport6);
  gtk_container_add (GTK_CONTAINER (scrolledwindow24), viewport6);
  gtk_viewport_set_shadow_type (GTK_VIEWPORT (viewport6), GTK_SHADOW_NONE);

  label156 = gtk_label_new (_("Please specify files/directories on the local filesystem to scan. Any files needed by jigdo templates that are found this way will not have to be downloaded again.\nNote that any directory to be scanned must contain single files which would otherwise be downloaded - in the case of CD images, you must supply the directory where the CD is mounted, NOT the name of an .iso file."));
  gtk_widget_show (label156);
  gtk_container_add (GTK_CONTAINER (viewport6), label156);
  gtk_label_set_justify (GTK_LABEL (label156), GTK_JUSTIFY_CENTER);
  gtk_label_set_line_wrap (GTK_LABEL (label156), TRUE);

  hbox53 = gtk_hbox_new (FALSE, 4);
  gtk_widget_show (hbox53);
  gtk_box_pack_start (GTK_BOX (pageReuse), hbox53, FALSE, FALSE, 0);

  reuse_path = gtk_entry_new ();
  gtk_widget_show (reuse_path);
  gtk_box_pack_start (GTK_BOX (hbox53), reuse_path, TRUE, TRUE, 0);

  reuse_pathSel = gtk_button_new_with_mnemonic (_("Browse..."));
  gtk_widget_show (reuse_pathSel);
  gtk_box_pack_start (GTK_BOX (hbox53), reuse_pathSel, FALSE, FALSE, 0);

  hbox58 = gtk_hbox_new (FALSE, 8);
  gtk_widget_show (hbox58);
  gtk_box_pack_start (GTK_BOX (pageReuse), hbox58, FALSE, FALSE, 0);

  reuse_scanButton = gtk_button_new ();
  gtk_widget_show (reuse_scanButton);
  gtk_box_pack_start (GTK_BOX (hbox58), reuse_scanButton, FALSE, FALSE, 0);
  GTK_WIDGET_SET_FLAGS (reuse_scanButton, GTK_CAN_DEFAULT);

  hbox59 = gtk_hbox_new (FALSE, 4);
  gtk_widget_show (hbox59);
  gtk_container_add (GTK_CONTAINER (reuse_scanButton), hbox59);
  gtk_container_set_border_width (GTK_CONTAINER (hbox59), 2);

  label161 = gtk_label_new (_("Start scan"));
  gtk_widget_show (label161);
  gtk_box_pack_start (GTK_BOX (hbox59), label161, FALSE, FALSE, 0);
  gtk_label_set_justify (GTK_LABEL (label161), GTK_JUSTIFY_CENTER);

  reuse_clearButton = gtk_button_new_with_mnemonic (_("Clear cache"));
  gtk_widget_show (reuse_clearButton);
  gtk_box_pack_start (GTK_BOX (hbox58), reuse_clearButton, FALSE, FALSE, 0);
  GTK_WIDGET_SET_FLAGS (reuse_clearButton, GTK_CAN_DEFAULT);

  pageReuseLabel = gtk_label_new (_("#Reuse"));
  gtk_widget_show (pageReuseLabel);
  gtk_notebook_set_tab_label (GTK_NOTEBOOK (invisibleNotebook), gtk_notebook_get_nth_page (GTK_NOTEBOOK (invisibleNotebook), 1), pageReuseLabel);
  gtk_label_set_justify (GTK_LABEL (pageReuseLabel), GTK_JUSTIFY_CENTER);

  pageDownload = gtk_vbox_new (FALSE, 0);
  gtk_widget_show (pageDownload);
  gtk_container_add (GTK_CONTAINER (invisibleNotebook), pageDownload);
  gtk_container_set_border_width (GTK_CONTAINER (pageDownload), 6);

  hbox96 = gtk_hbox_new (FALSE, 0);
  gtk_widget_show (hbox96);
  gtk_box_pack_end (GTK_BOX (pageDownload), hbox96, FALSE, FALSE, 0);

  _2lineshighlabelforsettingminimumheight = gtk_label_new (_("\n"));
  gtk_widget_show (_2lineshighlabelforsettingminimumheight);
  gtk_box_pack_start (GTK_BOX (hbox96), _2lineshighlabelforsettingminimumheight, FALSE, FALSE, 0);

  download_buttonInfo = gtk_label_new ("");
  gtk_widget_show (download_buttonInfo);
  gtk_box_pack_start (GTK_BOX (hbox96), download_buttonInfo, TRUE, TRUE, 0);
  gtk_label_set_use_markup (GTK_LABEL (download_buttonInfo), TRUE);
  gtk_label_set_justify (GTK_LABEL (download_buttonInfo), GTK_JUSTIFY_RIGHT);
  gtk_label_set_line_wrap (GTK_LABEL (download_buttonInfo), TRUE);
  gtk_misc_set_alignment (GTK_MISC (download_buttonInfo), 1, 0.5);
  gtk_misc_set_padding (GTK_MISC (download_buttonInfo), 12, 0);

  vbox42 = gtk_vbox_new (FALSE, 0);
  gtk_widget_show (vbox42);
  gtk_box_pack_end (GTK_BOX (hbox96), vbox42, FALSE, FALSE, 0);

  hbox83 = gtk_hbox_new (FALSE, 6);
  gtk_widget_show (hbox83);
  gtk_box_pack_end (GTK_BOX (vbox42), hbox83, FALSE, FALSE, 0);

  download_closeButton = gtk_button_new ();
  gtk_widget_show (download_closeButton);
  gtk_box_pack_end (GTK_BOX (hbox83), download_closeButton, FALSE, FALSE, 0);

  image14 = create_pixmap (window, "close.png");
  gtk_widget_show (image14);
  gtk_container_add (GTK_CONTAINER (download_closeButton), image14);

  download_restartButton = gtk_button_new ();
  gtk_widget_show (download_restartButton);
  gtk_box_pack_end (GTK_BOX (hbox83), download_restartButton, FALSE, FALSE, 0);

  image15 = create_pixmap (window, "restart.png");
  gtk_widget_show (image15);
  gtk_container_add (GTK_CONTAINER (download_restartButton), image15);

  label211 = gtk_label_new ("");
  gtk_widget_show (label211);
  gtk_box_pack_end (GTK_BOX (hbox83), label211, FALSE, FALSE, 3);

  download_stopButton = gtk_button_new ();
  gtk_widget_show (download_stopButton);
  gtk_box_pack_end (GTK_BOX (hbox83), download_stopButton, FALSE, FALSE, 0);

  image10 = create_pixmap (window, "stop.png");
  gtk_widget_show (image10);
  gtk_container_add (GTK_CONTAINER (download_stopButton), image10);

  download_pauseButton = gtk_button_new ();
  gtk_widget_show (download_pauseButton);
  gtk_box_pack_end (GTK_BOX (hbox83), download_pauseButton, FALSE, FALSE, 0);

  image11 = create_pixmap (window, "pause.png");
  gtk_widget_show (image11);
  gtk_container_add (GTK_CONTAINER (download_pauseButton), image11);

  download_startButton = gtk_button_new ();
  gtk_widget_show (download_startButton);
  gtk_box_pack_end (GTK_BOX (hbox83), download_startButton, FALSE, FALSE, 0);

  image13 = create_pixmap (window, "start.png");
  gtk_widget_show (image13);
  gtk_container_add (GTK_CONTAINER (download_startButton), image13);

  scrolledwindow23 = gtk_scrolled_window_new (NULL, NULL);
  gtk_widget_show (scrolledwindow23);
  gtk_box_pack_end (GTK_BOX (pageDownload), scrolledwindow23, FALSE, FALSE, 0);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolledwindow23), GTK_POLICY_AUTOMATIC, GTK_POLICY_NEVER);

  viewport5 = gtk_viewport_new (NULL, NULL);
  gtk_widget_show (viewport5);
  gtk_container_add (GTK_CONTAINER (scrolledwindow23), viewport5);
  gtk_viewport_set_shadow_type (GTK_VIEWPORT (viewport5), GTK_SHADOW_NONE);

  table5 = gtk_table_new (4, 2, FALSE);
  gtk_widget_show (table5);
  gtk_container_add (GTK_CONTAINER (viewport5), table5);
  gtk_container_set_border_width (GTK_CONTAINER (table5), 6);
  gtk_table_set_row_spacings (GTK_TABLE (table5), 4);
  gtk_table_set_col_spacings (GTK_TABLE (table5), 4);

  download_dest = gtk_label_new (_("#"));
  gtk_widget_show (download_dest);
  gtk_table_attach (GTK_TABLE (table5), download_dest, 1, 2, 1, 2,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);
  GTK_WIDGET_SET_FLAGS (download_dest, GTK_CAN_FOCUS);
  gtk_label_set_justify (GTK_LABEL (download_dest), GTK_JUSTIFY_CENTER);
  gtk_label_set_selectable (GTK_LABEL (download_dest), TRUE);
  gtk_misc_set_alignment (GTK_MISC (download_dest), 0, 0.5);

  download_progress = gtk_label_new (_("#"));
  gtk_widget_show (download_progress);
  gtk_table_attach (GTK_TABLE (table5), download_progress, 1, 2, 2, 3,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);
  GTK_WIDGET_SET_FLAGS (download_progress, GTK_CAN_FOCUS);
  gtk_label_set_justify (GTK_LABEL (download_progress), GTK_JUSTIFY_CENTER);
  gtk_label_set_selectable (GTK_LABEL (download_progress), TRUE);
  gtk_misc_set_alignment (GTK_MISC (download_progress), 0, 0.5);

  label151 = gtk_label_new (_("<b>Status  </b>"));
  gtk_widget_show (label151);
  gtk_table_attach (GTK_TABLE (table5), label151, 0, 1, 3, 4,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);
  gtk_label_set_use_markup (GTK_LABEL (label151), TRUE);
  gtk_label_set_justify (GTK_LABEL (label151), GTK_JUSTIFY_CENTER);
  gtk_misc_set_alignment (GTK_MISC (label151), 0, 0.5);

  download_status = gtk_label_new (_("#"));
  gtk_widget_show (download_status);
  gtk_table_attach (GTK_TABLE (table5), download_status, 1, 2, 3, 4,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);
  GTK_WIDGET_SET_FLAGS (download_status, GTK_CAN_FOCUS);
  gtk_label_set_justify (GTK_LABEL (download_status), GTK_JUSTIFY_CENTER);
  gtk_label_set_selectable (GTK_LABEL (download_status), TRUE);
  gtk_misc_set_alignment (GTK_MISC (download_status), 0, 0.5);

  download_URL = gtk_label_new (_("#"));
  gtk_widget_show (download_URL);
  gtk_table_attach (GTK_TABLE (table5), download_URL, 1, 2, 0, 1,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);
  GTK_WIDGET_SET_FLAGS (download_URL, GTK_CAN_FOCUS);
  gtk_label_set_justify (GTK_LABEL (download_URL), GTK_JUSTIFY_CENTER);
  gtk_label_set_selectable (GTK_LABEL (download_URL), TRUE);
  gtk_misc_set_alignment (GTK_MISC (download_URL), 0, 0.5);

  label131 = gtk_label_new (_("<b>URL  </b>"));
  gtk_widget_show (label131);
  gtk_table_attach (GTK_TABLE (table5), label131, 0, 1, 0, 1,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);
  gtk_label_set_use_markup (GTK_LABEL (label131), TRUE);
  gtk_label_set_justify (GTK_LABEL (label131), GTK_JUSTIFY_CENTER);
  gtk_misc_set_alignment (GTK_MISC (label131), 0, 0.5);

  label133 = gtk_label_new (_("<b>Saving to  </b>"));
  gtk_widget_show (label133);
  gtk_table_attach (GTK_TABLE (table5), label133, 0, 1, 1, 2,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);
  gtk_label_set_use_markup (GTK_LABEL (label133), TRUE);
  gtk_label_set_justify (GTK_LABEL (label133), GTK_JUSTIFY_CENTER);
  gtk_misc_set_alignment (GTK_MISC (label133), 0, 0.5);

  label149 = gtk_label_new (_("<b>Progress  </b>"));
  gtk_widget_show (label149);
  gtk_table_attach (GTK_TABLE (table5), label149, 0, 1, 2, 3,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);
  gtk_label_set_use_markup (GTK_LABEL (label149), TRUE);
  gtk_label_set_justify (GTK_LABEL (label149), GTK_JUSTIFY_CENTER);
  gtk_misc_set_alignment (GTK_MISC (label149), 0, 0.5);

  pageDownloadLabel = gtk_label_new (_("#Download"));
  gtk_widget_show (pageDownloadLabel);
  gtk_notebook_set_tab_label (GTK_NOTEBOOK (invisibleNotebook), gtk_notebook_get_nth_page (GTK_NOTEBOOK (invisibleNotebook), 2), pageDownloadLabel);
  gtk_label_set_justify (GTK_LABEL (pageDownloadLabel), GTK_JUSTIFY_CENTER);

  pageJigdo = gtk_notebook_new ();
  gtk_widget_show (pageJigdo);
  gtk_container_add (GTK_CONTAINER (invisibleNotebook), pageJigdo);

  vbox34 = gtk_vbox_new (FALSE, 4);
  gtk_widget_show (vbox34);
  gtk_container_add (GTK_CONTAINER (pageJigdo), vbox34);
  gtk_container_set_border_width (GTK_CONTAINER (vbox34), 4);

  scrolledwindow20 = gtk_scrolled_window_new (NULL, NULL);
  gtk_widget_show (scrolledwindow20);
  gtk_box_pack_start (GTK_BOX (vbox34), scrolledwindow20, TRUE, TRUE, 0);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolledwindow20), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);

  viewport2 = gtk_viewport_new (NULL, NULL);
  gtk_widget_show (viewport2);
  gtk_container_add (GTK_CONTAINER (scrolledwindow20), viewport2);
  gtk_viewport_set_shadow_type (GTK_VIEWPORT (viewport2), GTK_SHADOW_NONE);

  jigdo_InfoVbox = gtk_vbox_new (FALSE, 8);
  gtk_widget_show (jigdo_InfoVbox);
  gtk_container_add (GTK_CONTAINER (viewport2), jigdo_InfoVbox);

  label222 = gtk_label_new (_("<b>Note: Even though some data is already being downloaded, you\nneed to complete step 2 before the actual file download can start!</b>"));
  gtk_widget_show (label222);
  gtk_box_pack_start (GTK_BOX (jigdo_InfoVbox), label222, FALSE, FALSE, 0);
  gtk_label_set_use_markup (GTK_LABEL (label222), TRUE);
  gtk_label_set_justify (GTK_LABEL (label222), GTK_JUSTIFY_CENTER);

  hseparator5 = gtk_hseparator_new ();
  gtk_widget_show (hseparator5);
  gtk_box_pack_start (GTK_BOX (jigdo_InfoVbox), hseparator5, FALSE, TRUE, 0);

  jigdo_ShortInfo = gtk_label_new (_("#<b>ShortInfo</b>"));
  gtk_widget_show (jigdo_ShortInfo);
  gtk_box_pack_start (GTK_BOX (jigdo_InfoVbox), jigdo_ShortInfo, FALSE, FALSE, 0);
  GTK_WIDGET_SET_FLAGS (jigdo_ShortInfo, GTK_CAN_FOCUS);
  gtk_label_set_use_markup (GTK_LABEL (jigdo_ShortInfo), TRUE);
  gtk_label_set_selectable (GTK_LABEL (jigdo_ShortInfo), TRUE);
  gtk_misc_set_alignment (GTK_MISC (jigdo_ShortInfo), 7.45058e-09, 0.5);

  jigdo_Info = gtk_label_new (_("#Info\nx\nx\nx"));
  gtk_widget_show (jigdo_Info);
  gtk_box_pack_start (GTK_BOX (jigdo_InfoVbox), jigdo_Info, FALSE, TRUE, 0);
  GTK_WIDGET_SET_FLAGS (jigdo_Info, GTK_CAN_FOCUS);
  gtk_label_set_use_markup (GTK_LABEL (jigdo_Info), TRUE);
  gtk_label_set_line_wrap (GTK_LABEL (jigdo_Info), TRUE);
  gtk_label_set_selectable (GTK_LABEL (jigdo_Info), TRUE);

  hbuttonbox9 = gtk_hbutton_box_new ();
  gtk_widget_show (hbuttonbox9);
  gtk_box_pack_start (GTK_BOX (vbox34), hbuttonbox9, FALSE, TRUE, 0);
  gtk_button_box_set_layout (GTK_BUTTON_BOX (hbuttonbox9), GTK_BUTTONBOX_END);
  gtk_box_set_spacing (GTK_BOX (hbuttonbox9), 6);

  button119 = gtk_button_new ();
  gtk_widget_show (button119);
  gtk_container_add (GTK_CONTAINER (hbuttonbox9), button119);
  GTK_WIDGET_SET_FLAGS (button119, GTK_CAN_DEFAULT);

  alignment3 = gtk_alignment_new (0.5, 0.5, 0, 0);
  gtk_widget_show (alignment3);
  gtk_container_add (GTK_CONTAINER (button119), alignment3);

  hbox85 = gtk_hbox_new (FALSE, 2);
  gtk_widget_show (hbox85);
  gtk_container_add (GTK_CONTAINER (alignment3), hbox85);

  image16 = gtk_image_new_from_stock ("gtk-go-forward", GTK_ICON_SIZE_BUTTON);
  gtk_widget_show (image16);
  gtk_box_pack_end (GTK_BOX (hbox85), image16, TRUE, TRUE, 0);

  label223 = gtk_label_new_with_mnemonic (_("_Next"));
  gtk_widget_show (label223);
  gtk_box_pack_start (GTK_BOX (hbox85), label223, FALSE, FALSE, 0);
  gtk_label_set_use_markup (GTK_LABEL (label223), TRUE);

  label213 = gtk_label_new_with_mnemonic (_("<b>_1</b>: Information"));
  gtk_widget_show (label213);
  gtk_notebook_set_tab_label (GTK_NOTEBOOK (pageJigdo), gtk_notebook_get_nth_page (GTK_NOTEBOOK (pageJigdo), 0), label213);
  gtk_label_set_use_markup (GTK_LABEL (label213), TRUE);
  gtk_misc_set_alignment (GTK_MISC (label213), 0, 0.5);

  vbox36 = gtk_vbox_new (FALSE, 12);
  gtk_container_add (GTK_CONTAINER (pageJigdo), vbox36);
  gtk_container_set_border_width (GTK_CONTAINER (vbox36), 4);

  vbox39 = gtk_vbox_new (FALSE, 0);
  gtk_widget_show (vbox39);
  gtk_box_pack_start (GTK_BOX (vbox36), vbox39, TRUE, TRUE, 0);

  label229 = gtk_label_new (_("<b>Reuse local files for generation of %1:</b>"));
  gtk_widget_show (label229);
  gtk_box_pack_start (GTK_BOX (vbox39), label229, FALSE, FALSE, 0);
  gtk_label_set_use_markup (GTK_LABEL (label229), TRUE);
  gtk_misc_set_alignment (GTK_MISC (label229), 7.45058e-09, 0.5);

  label230 = gtk_label_new (_("UNIMPLEMENTED\njust click Next..."));
  gtk_widget_show (label230);
  gtk_box_pack_start (GTK_BOX (vbox39), label230, TRUE, TRUE, 0);
  gtk_label_set_justify (GTK_LABEL (label230), GTK_JUSTIFY_CENTER);

  hbox88 = gtk_hbox_new (FALSE, 4);
  gtk_widget_show (hbox88);
  gtk_box_pack_start (GTK_BOX (vbox36), hbox88, FALSE, TRUE, 0);

  entry23 = gtk_entry_new ();
  gtk_widget_show (entry23);
  gtk_box_pack_start (GTK_BOX (hbox88), entry23, TRUE, TRUE, 0);
  gtk_widget_set_sensitive (entry23, FALSE);

  alignment6 = gtk_alignment_new (0.5, 0.5, 1, 0);
  gtk_widget_show (alignment6);
  gtk_box_pack_start (GTK_BOX (hbox88), alignment6, FALSE, FALSE, 0);

  button127 = gtk_button_new_with_mnemonic (_("..."));
  gtk_widget_show (button127);
  gtk_container_add (GTK_CONTAINER (alignment6), button127);
  gtk_widget_set_sensitive (button127, FALSE);

  button128 = gtk_button_new ();
  gtk_widget_show (button128);
  gtk_box_pack_start (GTK_BOX (hbox88), button128, FALSE, FALSE, 0);
  gtk_widget_set_sensitive (button128, FALSE);
  GTK_WIDGET_SET_FLAGS (button128, GTK_CAN_DEFAULT);

  alignment5 = gtk_alignment_new (0.5, 0.5, 0, 0);
  gtk_widget_show (alignment5);
  gtk_container_add (GTK_CONTAINER (button128), alignment5);

  hbox89 = gtk_hbox_new (FALSE, 2);
  gtk_widget_show (hbox89);
  gtk_container_add (GTK_CONTAINER (alignment5), hbox89);

  image23 = gtk_image_new_from_stock ("gtk-ok", GTK_ICON_SIZE_BUTTON);
  gtk_widget_show (image23);
  gtk_box_pack_start (GTK_BOX (hbox89), image23, TRUE, TRUE, 0);

  label227 = gtk_label_new_with_mnemonic (_("_Start scan"));
  gtk_widget_show (label227);
  gtk_box_pack_start (GTK_BOX (hbox89), label227, FALSE, FALSE, 0);

  hbuttonbox10 = gtk_hbutton_box_new ();
  gtk_widget_show (hbuttonbox10);
  gtk_box_pack_start (GTK_BOX (vbox36), hbuttonbox10, FALSE, TRUE, 0);
  gtk_button_box_set_layout (GTK_BUTTON_BOX (hbuttonbox10), GTK_BUTTONBOX_END);
  gtk_box_set_spacing (GTK_BOX (hbuttonbox10), 6);

  button120 = gtk_button_new_from_stock ("gtk-go-back");
  gtk_widget_show (button120);
  gtk_container_add (GTK_CONTAINER (hbuttonbox10), button120);
  GTK_WIDGET_SET_FLAGS (button120, GTK_CAN_DEFAULT);

  button121 = gtk_button_new ();
  gtk_widget_show (button121);
  gtk_container_add (GTK_CONTAINER (hbuttonbox10), button121);
  GTK_WIDGET_SET_FLAGS (button121, GTK_CAN_DEFAULT);

  alignment4 = gtk_alignment_new (0.5, 0.5, 0, 0);
  gtk_widget_show (alignment4);
  gtk_container_add (GTK_CONTAINER (button121), alignment4);

  hbox86 = gtk_hbox_new (FALSE, 2);
  gtk_widget_show (hbox86);
  gtk_container_add (GTK_CONTAINER (alignment4), hbox86);

  image17 = gtk_image_new_from_stock ("gtk-go-forward", GTK_ICON_SIZE_BUTTON);
  gtk_widget_show (image17);
  gtk_box_pack_end (GTK_BOX (hbox86), image17, TRUE, TRUE, 0);

  label224 = gtk_label_new_with_mnemonic (_("_Next"));
  gtk_widget_show (label224);
  gtk_box_pack_start (GTK_BOX (hbox86), label224, FALSE, FALSE, 0);

  label217 = gtk_label_new (_("<b>?</b>: Reuse files"));
  gtk_widget_show (label217);
  gtk_notebook_set_tab_label (GTK_NOTEBOOK (pageJigdo), gtk_notebook_get_nth_page (GTK_NOTEBOOK (pageJigdo), 1), label217);
  gtk_label_set_use_markup (GTK_LABEL (label217), TRUE);
  gtk_misc_set_alignment (GTK_MISC (label217), 0, 0.5);

  vbox40 = gtk_vbox_new (FALSE, 12);
  gtk_widget_show (vbox40);
  gtk_container_add (GTK_CONTAINER (pageJigdo), vbox40);
  gtk_container_set_border_width (GTK_CONTAINER (vbox40), 4);

  vbox41 = gtk_vbox_new (FALSE, 4);
  gtk_widget_show (vbox41);
  gtk_box_pack_start (GTK_BOX (vbox40), vbox41, TRUE, TRUE, 0);

  label239 = gtk_label_new (_("<b>Select servers to download data from:</b>"));
  gtk_widget_show (label239);
  gtk_box_pack_start (GTK_BOX (vbox41), label239, FALSE, FALSE, 0);
  gtk_label_set_use_markup (GTK_LABEL (label239), TRUE);
  gtk_misc_set_alignment (GTK_MISC (label239), 7.45058e-09, 0.5);

  radiobutton1 = gtk_radio_button_new_with_mnemonic (NULL, _("_Automatic selection: Use \"netselect\" to find the best servers"));
  gtk_box_pack_start (GTK_BOX (vbox41), radiobutton1, FALSE, FALSE, 0);
  gtk_radio_button_set_group (GTK_RADIO_BUTTON (radiobutton1), radiobutton1_group);
  radiobutton1_group = gtk_radio_button_get_group (GTK_RADIO_BUTTON (radiobutton1));

  hbox93 = gtk_hbox_new (FALSE, 0);
  gtk_widget_show (hbox93);
  gtk_box_pack_start (GTK_BOX (vbox41), hbox93, FALSE, FALSE, 0);

  radiobutton2 = gtk_radio_button_new_with_mnemonic (NULL, _("_Simple selection, by country name: "));
  gtk_widget_show (radiobutton2);
  gtk_box_pack_start (GTK_BOX (hbox93), radiobutton2, FALSE, FALSE, 0);
  gtk_tooltips_set_tip (tooltips, radiobutton2, _("Select the country you live in. jigdo will then try out all available servers on your continent. Sooner or later, it should end up using a server that is fast for you."), NULL);
  gtk_radio_button_set_group (GTK_RADIO_BUTTON (radiobutton2), radiobutton1_group);
  radiobutton1_group = gtk_radio_button_get_group (GTK_RADIO_BUTTON (radiobutton2));

  combo1 = gtk_combo_new ();
  g_object_set_data (G_OBJECT (GTK_COMBO (combo1)->popwin),
                     "GladeParentKey", combo1);
  gtk_widget_show (combo1);
  gtk_box_pack_start (GTK_BOX (hbox93), combo1, TRUE, TRUE, 0);

  combo_entry1 = GTK_COMBO (combo1)->entry;
  gtk_widget_show (combo_entry1);
  gtk_editable_set_editable (GTK_EDITABLE (combo_entry1), FALSE);
  gtk_entry_set_text (GTK_ENTRY (combo_entry1), _("Select the country you live in â\206\222"));

  hbox94 = gtk_hbox_new (FALSE, 0);
  gtk_widget_show (hbox94);
  gtk_box_pack_start (GTK_BOX (vbox41), hbox94, TRUE, TRUE, 0);

  alignment10 = gtk_alignment_new (0.5, 0, 1, 0);
  gtk_widget_show (alignment10);
  gtk_box_pack_start (GTK_BOX (hbox94), alignment10, FALSE, FALSE, 0);

  radiobutton3 = gtk_radio_button_new_with_mnemonic (NULL, _("Ad_vanced: "));
  gtk_widget_show (radiobutton3);
  gtk_container_add (GTK_CONTAINER (alignment10), radiobutton3);
  gtk_tooltips_set_tip (tooltips, radiobutton3, _("Select or enter server URLs for the different types of servers that the .jigdo file references. You can enter several URLs for each type of server - in this case, jigdo will try to find the fastest."), NULL);
  gtk_radio_button_set_group (GTK_RADIO_BUTTON (radiobutton3), radiobutton1_group);
  radiobutton1_group = gtk_radio_button_get_group (GTK_RADIO_BUTTON (radiobutton3));

  scrolledwindow21 = gtk_scrolled_window_new (NULL, NULL);
  gtk_widget_show (scrolledwindow21);
  gtk_box_pack_start (GTK_BOX (hbox94), scrolledwindow21, TRUE, TRUE, 0);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolledwindow21), GTK_POLICY_NEVER, GTK_POLICY_AUTOMATIC);

  viewport3 = gtk_viewport_new (NULL, NULL);
  gtk_widget_show (viewport3);
  gtk_container_add (GTK_CONTAINER (scrolledwindow21), viewport3);
  gtk_viewport_set_shadow_type (GTK_VIEWPORT (viewport3), GTK_SHADOW_NONE);

  table8 = gtk_table_new (3, 2, FALSE);
  gtk_widget_show (table8);
  gtk_container_add (GTK_CONTAINER (viewport3), table8);
  gtk_table_set_row_spacings (GTK_TABLE (table8), 2);
  gtk_table_set_col_spacings (GTK_TABLE (table8), 2);

  label243 = gtk_label_new (_("<b>Type of server:   </b>"));
  gtk_widget_show (label243);
  gtk_table_attach (GTK_TABLE (table8), label243, 0, 1, 0, 1,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);
  gtk_label_set_use_markup (GTK_LABEL (label243), TRUE);
  gtk_misc_set_alignment (GTK_MISC (label243), 0, 0.5);

  label244 = gtk_label_new (_("<b>URL(s) for that server:</b> (space-separated list, preferred URLs first)"));
  gtk_widget_show (label244);
  gtk_table_attach (GTK_TABLE (table8), label244, 1, 2, 0, 1,
                    (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);
  gtk_label_set_use_markup (GTK_LABEL (label244), TRUE);
  gtk_misc_set_alignment (GTK_MISC (label244), 0, 0.5);

  label245 = gtk_label_new (_("Debian"));
  gtk_widget_show (label245);
  gtk_table_attach (GTK_TABLE (table8), label245, 0, 1, 1, 2,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);
  gtk_misc_set_alignment (GTK_MISC (label245), 0, 0.5);

  label246 = gtk_label_new (_("Non-US"));
  gtk_widget_show (label246);
  gtk_table_attach (GTK_TABLE (table8), label246, 0, 1, 2, 3,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);
  gtk_misc_set_alignment (GTK_MISC (label246), 0, 0.5);

  combo2 = gtk_combo_new ();
  g_object_set_data (G_OBJECT (GTK_COMBO (combo2)->popwin),
                     "GladeParentKey", combo2);
  gtk_widget_show (combo2);
  gtk_table_attach (GTK_TABLE (table8), combo2, 1, 2, 1, 2,
                    (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);

  combo_entry2 = GTK_COMBO (combo2)->entry;
  gtk_widget_show (combo_entry2);

  combo3 = gtk_combo_new ();
  g_object_set_data (G_OBJECT (GTK_COMBO (combo3)->popwin),
                     "GladeParentKey", combo3);
  gtk_widget_show (combo3);
  gtk_table_attach (GTK_TABLE (table8), combo3, 1, 2, 2, 3,
                    (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);

  combo_entry3 = GTK_COMBO (combo3)->entry;
  gtk_widget_show (combo_entry3);

  hbuttonbox11 = gtk_hbutton_box_new ();
  gtk_widget_show (hbuttonbox11);
  gtk_box_pack_start (GTK_BOX (vbox40), hbuttonbox11, FALSE, TRUE, 0);
  gtk_button_box_set_layout (GTK_BUTTON_BOX (hbuttonbox11), GTK_BUTTONBOX_END);
  gtk_box_set_spacing (GTK_BOX (hbuttonbox11), 6);

  button131 = gtk_button_new_from_stock ("gtk-go-back");
  gtk_widget_show (button131);
  gtk_container_add (GTK_CONTAINER (hbuttonbox11), button131);
  GTK_WIDGET_SET_FLAGS (button131, GTK_CAN_DEFAULT);

  button132 = gtk_button_new ();
  gtk_widget_show (button132);
  gtk_container_add (GTK_CONTAINER (hbuttonbox11), button132);
  GTK_WIDGET_SET_FLAGS (button132, GTK_CAN_DEFAULT);

  alignment9 = gtk_alignment_new (0.5, 0.5, 0, 0);
  gtk_widget_show (alignment9);
  gtk_container_add (GTK_CONTAINER (button132), alignment9);

  hbox92 = gtk_hbox_new (FALSE, 2);
  gtk_widget_show (hbox92);
  gtk_container_add (GTK_CONTAINER (alignment9), hbox92);

  image25 = gtk_image_new_from_stock ("gtk-go-forward", GTK_ICON_SIZE_BUTTON);
  gtk_widget_show (image25);
  gtk_box_pack_end (GTK_BOX (hbox92), image25, TRUE, TRUE, 0);

  label242 = gtk_label_new_with_mnemonic (_("<b>_OK, start download!</b>"));
  gtk_widget_show (label242);
  gtk_box_pack_start (GTK_BOX (hbox92), label242, FALSE, FALSE, 0);
  gtk_label_set_use_markup (GTK_LABEL (label242), TRUE);

  label218 = gtk_label_new_with_mnemonic (_("<b>_2</b>: Select servers"));
  gtk_widget_show (label218);
  gtk_notebook_set_tab_label (GTK_NOTEBOOK (pageJigdo), gtk_notebook_get_nth_page (GTK_NOTEBOOK (pageJigdo), 2), label218);
  gtk_label_set_use_markup (GTK_LABEL (label218), TRUE);
  gtk_misc_set_alignment (GTK_MISC (label218), 0, 0.5);

  vbox37 = gtk_vbox_new (FALSE, 0);
  gtk_widget_show (vbox37);
  gtk_container_add (GTK_CONTAINER (pageJigdo), vbox37);
  gtk_container_set_border_width (GTK_CONTAINER (vbox37), 4);

  hbox97 = gtk_hbox_new (FALSE, 0);
  gtk_widget_show (hbox97);
  gtk_box_pack_end (GTK_BOX (vbox37), hbox97, FALSE, FALSE, 0);

  _2lineshighforsettingminheight = gtk_label_new (_("\n"));
  gtk_widget_show (_2lineshighforsettingminheight);
  gtk_box_pack_start (GTK_BOX (hbox97), _2lineshighforsettingminheight, FALSE, FALSE, 0);

  jigdo_buttonInfo = gtk_label_new ("");
  gtk_widget_show (jigdo_buttonInfo);
  gtk_box_pack_start (GTK_BOX (hbox97), jigdo_buttonInfo, TRUE, TRUE, 0);
  gtk_label_set_use_markup (GTK_LABEL (jigdo_buttonInfo), TRUE);
  gtk_label_set_justify (GTK_LABEL (jigdo_buttonInfo), GTK_JUSTIFY_RIGHT);
  gtk_label_set_line_wrap (GTK_LABEL (jigdo_buttonInfo), TRUE);
  gtk_misc_set_alignment (GTK_MISC (jigdo_buttonInfo), 1, 0.5);
  gtk_misc_set_padding (GTK_MISC (jigdo_buttonInfo), 12, 0);

  vbox43 = gtk_vbox_new (FALSE, 0);
  gtk_widget_show (vbox43);
  gtk_box_pack_end (GTK_BOX (hbox97), vbox43, FALSE, FALSE, 0);

  hbox98 = gtk_hbox_new (FALSE, 6);
  gtk_widget_show (hbox98);
  gtk_box_pack_end (GTK_BOX (vbox43), hbox98, FALSE, FALSE, 0);

  jigdo_closeButton = gtk_button_new ();
  gtk_widget_show (jigdo_closeButton);
  gtk_box_pack_end (GTK_BOX (hbox98), jigdo_closeButton, FALSE, FALSE, 0);

  image28 = create_pixmap (window, "close.png");
  gtk_widget_show (image28);
  gtk_container_add (GTK_CONTAINER (jigdo_closeButton), image28);

  jigdo_restartButton = gtk_button_new ();
  gtk_widget_show (jigdo_restartButton);
  gtk_box_pack_end (GTK_BOX (hbox98), jigdo_restartButton, FALSE, FALSE, 0);

  image29 = create_pixmap (window, "restart.png");
  gtk_widget_show (image29);
  gtk_container_add (GTK_CONTAINER (jigdo_restartButton), image29);

  label251 = gtk_label_new ("");
  gtk_widget_show (label251);
  gtk_box_pack_end (GTK_BOX (hbox98), label251, FALSE, FALSE, 3);

  jigdo_stopButton = gtk_button_new ();
  gtk_widget_show (jigdo_stopButton);
  gtk_box_pack_end (GTK_BOX (hbox98), jigdo_stopButton, FALSE, FALSE, 0);

  image30 = create_pixmap (window, "stop.png");
  gtk_widget_show (image30);
  gtk_container_add (GTK_CONTAINER (jigdo_stopButton), image30);

  jigdo_pauseButton = gtk_button_new ();
  gtk_widget_show (jigdo_pauseButton);
  gtk_box_pack_end (GTK_BOX (hbox98), jigdo_pauseButton, FALSE, FALSE, 0);

  image31 = create_pixmap (window, "pause.png");
  gtk_widget_show (image31);
  gtk_container_add (GTK_CONTAINER (jigdo_pauseButton), image31);

  jigdo_startButton = gtk_button_new ();
  gtk_widget_show (jigdo_startButton);
  gtk_box_pack_end (GTK_BOX (hbox98), jigdo_startButton, FALSE, FALSE, 0);

  image32 = create_pixmap (window, "start.png");
  gtk_widget_show (image32);
  gtk_container_add (GTK_CONTAINER (jigdo_startButton), image32);

  scrolledwindow22 = gtk_scrolled_window_new (NULL, NULL);
  gtk_widget_show (scrolledwindow22);
  gtk_box_pack_end (GTK_BOX (vbox37), scrolledwindow22, FALSE, FALSE, 0);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolledwindow22), GTK_POLICY_AUTOMATIC, GTK_POLICY_NEVER);

  viewport4 = gtk_viewport_new (NULL, NULL);
  gtk_widget_show (viewport4);
  gtk_container_add (GTK_CONTAINER (scrolledwindow22), viewport4);
  gtk_viewport_set_shadow_type (GTK_VIEWPORT (viewport4), GTK_SHADOW_NONE);

  table7 = gtk_table_new (4, 2, FALSE);
  gtk_widget_show (table7);
  gtk_container_add (GTK_CONTAINER (viewport4), table7);
  gtk_container_set_border_width (GTK_CONTAINER (table7), 6);
  gtk_table_set_row_spacings (GTK_TABLE (table7), 4);
  gtk_table_set_col_spacings (GTK_TABLE (table7), 4);

  label233 = gtk_label_new (_("<b>Saving to  </b>"));
  gtk_widget_show (label233);
  gtk_table_attach (GTK_TABLE (table7), label233, 0, 1, 1, 2,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);
  gtk_label_set_use_markup (GTK_LABEL (label233), TRUE);
  gtk_label_set_justify (GTK_LABEL (label233), GTK_JUSTIFY_CENTER);
  gtk_misc_set_alignment (GTK_MISC (label233), 0, 0.5);

  label235 = gtk_label_new (_("<b>Progress  </b>"));
  gtk_widget_show (label235);
  gtk_table_attach (GTK_TABLE (table7), label235, 0, 1, 2, 3,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);
  gtk_label_set_use_markup (GTK_LABEL (label235), TRUE);
  gtk_label_set_justify (GTK_LABEL (label235), GTK_JUSTIFY_CENTER);
  gtk_misc_set_alignment (GTK_MISC (label235), 0, 0.5);

  label237 = gtk_label_new (_("<b>Status  </b>"));
  gtk_widget_show (label237);
  gtk_table_attach (GTK_TABLE (table7), label237, 0, 1, 3, 4,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);
  gtk_label_set_use_markup (GTK_LABEL (label237), TRUE);
  gtk_label_set_justify (GTK_LABEL (label237), GTK_JUSTIFY_CENTER);
  gtk_misc_set_alignment (GTK_MISC (label237), 0, 0.5);

  label231 = gtk_label_new (_("<b>URL  </b>"));
  gtk_widget_show (label231);
  gtk_table_attach (GTK_TABLE (table7), label231, 0, 1, 0, 1,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);
  gtk_label_set_use_markup (GTK_LABEL (label231), TRUE);
  gtk_label_set_justify (GTK_LABEL (label231), GTK_JUSTIFY_CENTER);
  gtk_misc_set_alignment (GTK_MISC (label231), 0, 0.5);

  jigdo_status = gtk_label_new (_("#"));
  gtk_widget_show (jigdo_status);
  gtk_table_attach (GTK_TABLE (table7), jigdo_status, 1, 2, 3, 4,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);
  GTK_WIDGET_SET_FLAGS (jigdo_status, GTK_CAN_FOCUS);
  gtk_label_set_justify (GTK_LABEL (jigdo_status), GTK_JUSTIFY_CENTER);
  gtk_label_set_selectable (GTK_LABEL (jigdo_status), TRUE);
  gtk_misc_set_alignment (GTK_MISC (jigdo_status), 0, 0.5);

  jigdo_progress = gtk_label_new (_("#"));
  gtk_widget_show (jigdo_progress);
  gtk_table_attach (GTK_TABLE (table7), jigdo_progress, 1, 2, 2, 3,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);
  GTK_WIDGET_SET_FLAGS (jigdo_progress, GTK_CAN_FOCUS);
  gtk_label_set_justify (GTK_LABEL (jigdo_progress), GTK_JUSTIFY_CENTER);
  gtk_label_set_selectable (GTK_LABEL (jigdo_progress), TRUE);
  gtk_misc_set_alignment (GTK_MISC (jigdo_progress), 0, 0.5);

  jigdo_dest = gtk_label_new (_("#"));
  gtk_widget_show (jigdo_dest);
  gtk_table_attach (GTK_TABLE (table7), jigdo_dest, 1, 2, 1, 2,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);
  GTK_WIDGET_SET_FLAGS (jigdo_dest, GTK_CAN_FOCUS);
  gtk_label_set_justify (GTK_LABEL (jigdo_dest), GTK_JUSTIFY_CENTER);
  gtk_label_set_selectable (GTK_LABEL (jigdo_dest), TRUE);
  gtk_misc_set_alignment (GTK_MISC (jigdo_dest), 0, 0.5);

  jigdo_URL = gtk_label_new (_("#"));
  gtk_widget_show (jigdo_URL);
  gtk_table_attach (GTK_TABLE (table7), jigdo_URL, 1, 2, 0, 1,
                    (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);
  GTK_WIDGET_SET_FLAGS (jigdo_URL, GTK_CAN_FOCUS);
  gtk_label_set_justify (GTK_LABEL (jigdo_URL), GTK_JUSTIFY_CENTER);
  gtk_label_set_selectable (GTK_LABEL (jigdo_URL), TRUE);
  gtk_misc_set_alignment (GTK_MISC (jigdo_URL), 0, 0.5);

  label219 = gtk_label_new_with_mnemonic (_("<b>_3</b>: Download"));
  gtk_widget_show (label219);
  gtk_notebook_set_tab_label (GTK_NOTEBOOK (pageJigdo), gtk_notebook_get_nth_page (GTK_NOTEBOOK (pageJigdo), 3), label219);
  gtk_label_set_use_markup (GTK_LABEL (label219), TRUE);
  gtk_misc_set_alignment (GTK_MISC (label219), 0, 0.5);

  pageJigdoLabel = gtk_label_new (_("#Jigdo"));
  gtk_widget_show (pageJigdoLabel);
  gtk_notebook_set_tab_label (GTK_NOTEBOOK (invisibleNotebook), gtk_notebook_get_nth_page (GTK_NOTEBOOK (invisibleNotebook), 3), pageJigdoLabel);

  progressScroll = gtk_scrolled_window_new (NULL, NULL);
  gtk_widget_show (progressScroll);
  gtk_paned_pack2 (GTK_PANED (windowPaned), progressScroll, TRUE, TRUE);
  GTK_WIDGET_UNSET_FLAGS (progressScroll, GTK_CAN_FOCUS);
  gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (progressScroll), GTK_SHADOW_IN);

  jobs = gtk_tree_view_new ();
  gtk_widget_show (jobs);
  gtk_container_add (GTK_CONTAINER (progressScroll), jobs);
  gtk_tree_view_set_rules_hint (GTK_TREE_VIEW (jobs), TRUE);

  g_signal_connect ((gpointer) window, "delete_event",
                    G_CALLBACK (on_window_delete_event),
                    NULL);
  g_signal_connect_swapped ((gpointer) toolbarOpen, "clicked",
                            G_CALLBACK (setNotebookPage),
                            GTK_OBJECT (pageOpen));
  g_signal_connect_swapped ((gpointer) toolbarReuse, "clicked",
                            G_CALLBACK (setNotebookPage),
                            GTK_OBJECT (pageReuse));
  g_signal_connect ((gpointer) toolbarExit, "clicked",
                    G_CALLBACK (on_toolbarExit_clicked),
                    NULL);
  g_signal_connect ((gpointer) aboutJigdoButton, "clicked",
                    G_CALLBACK (on_aboutJigdoButton_clicked),
                    NULL);
  g_signal_connect ((gpointer) openButton, "clicked",
                    G_CALLBACK (on_openButton_clicked),
                    NULL);
  g_signal_connect ((gpointer) open_URL, "activate",
                    G_CALLBACK (on_openButton_clicked),
                    NULL);
  g_signal_connect ((gpointer) open_dest, "activate",
                    G_CALLBACK (on_openButton_clicked),
                    NULL);
  g_signal_connect ((gpointer) download_closeButton, "enter",
                    G_CALLBACK (on_download_closeButton_enter),
                    NULL);
  g_signal_connect ((gpointer) download_closeButton, "leave",
                    G_CALLBACK (on_download_button_leave),
                    NULL);
  g_signal_connect ((gpointer) download_closeButton, "clicked",
                    G_CALLBACK (on_download_closeButton_clicked),
                    NULL);
  g_signal_connect ((gpointer) download_restartButton, "enter",
                    G_CALLBACK (on_download_restartButton_enter),
                    NULL);
  g_signal_connect ((gpointer) download_restartButton, "leave",
                    G_CALLBACK (on_download_button_leave),
                    NULL);
  g_signal_connect ((gpointer) download_restartButton, "clicked",
                    G_CALLBACK (on_download_restartButton_clicked),
                    NULL);
  g_signal_connect ((gpointer) download_stopButton, "enter",
                    G_CALLBACK (on_download_stopButton_enter),
                    NULL);
  g_signal_connect ((gpointer) download_stopButton, "leave",
                    G_CALLBACK (on_download_button_leave),
                    NULL);
  g_signal_connect ((gpointer) download_stopButton, "clicked",
                    G_CALLBACK (on_download_stopButton_clicked),
                    NULL);
  g_signal_connect ((gpointer) download_pauseButton, "enter",
                    G_CALLBACK (on_download_pauseButton_enter),
                    NULL);
  g_signal_connect ((gpointer) download_pauseButton, "leave",
                    G_CALLBACK (on_download_button_leave),
                    NULL);
  g_signal_connect ((gpointer) download_pauseButton, "clicked",
                    G_CALLBACK (on_download_pauseButton_clicked),
                    NULL);
  g_signal_connect ((gpointer) download_startButton, "enter",
                    G_CALLBACK (on_download_startButton_enter),
                    NULL);
  g_signal_connect ((gpointer) download_startButton, "leave",
                    G_CALLBACK (on_download_button_leave),
                    NULL);
  g_signal_connect ((gpointer) download_startButton, "clicked",
                    G_CALLBACK (on_download_startButton_clicked),
                    NULL);
  g_signal_connect ((gpointer) entry23, "activate",
                    G_CALLBACK (on_openButton_clicked),
                    NULL);
  g_signal_connect ((gpointer) jigdo_closeButton, "enter",
                    G_CALLBACK (on_jigdo_closeButton_enter),
                    NULL);
  g_signal_connect ((gpointer) jigdo_closeButton, "leave",
                    G_CALLBACK (on_jigdo_button_leave),
                    NULL);
  g_signal_connect ((gpointer) jigdo_closeButton, "clicked",
                    G_CALLBACK (on_jigdo_closeButton_clicked),
                    NULL);
  g_signal_connect ((gpointer) jigdo_restartButton, "enter",
                    G_CALLBACK (on_jigdo_restartButton_enter),
                    NULL);
  g_signal_connect ((gpointer) jigdo_restartButton, "leave",
                    G_CALLBACK (on_jigdo_button_leave),
                    NULL);
  g_signal_connect ((gpointer) jigdo_restartButton, "clicked",
                    G_CALLBACK (on_jigdo_restartButton_clicked),
                    NULL);
  g_signal_connect ((gpointer) jigdo_stopButton, "enter",
                    G_CALLBACK (on_jigdo_stopButton_enter),
                    NULL);
  g_signal_connect ((gpointer) jigdo_stopButton, "leave",
                    G_CALLBACK (on_jigdo_button_leave),
                    NULL);
  g_signal_connect ((gpointer) jigdo_stopButton, "clicked",
                    G_CALLBACK (on_jigdo_stopButton_clicked),
                    NULL);
  g_signal_connect ((gpointer) jigdo_pauseButton, "enter",
                    G_CALLBACK (on_jigdo_pauseButton_enter),
                    NULL);
  g_signal_connect ((gpointer) jigdo_pauseButton, "leave",
                    G_CALLBACK (on_jigdo_button_leave),
                    NULL);
  g_signal_connect ((gpointer) jigdo_pauseButton, "clicked",
                    G_CALLBACK (on_jigdo_pauseButton_clicked),
                    NULL);
  g_signal_connect ((gpointer) jigdo_startButton, "enter",
                    G_CALLBACK (on_jigdo_startButton_enter),
                    NULL);
  g_signal_connect ((gpointer) jigdo_startButton, "leave",
                    G_CALLBACK (on_jigdo_button_leave),
                    NULL);
  g_signal_connect ((gpointer) jigdo_startButton, "clicked",
                    G_CALLBACK (on_jigdo_startButton_clicked),
                    NULL);

  atko = gtk_widget_get_accessible (aboutJigdoLogo);
  atk_object_set_description (atko, _("Jigdo logo"));


  /* Store pointers to all widgets, for use by lookup_widget(). */
  GLADE_HOOKUP_OBJECT_NO_REF (window, window, "window");
  GLADE_HOOKUP_OBJECT (window, windowVbox, "windowVbox");
  GLADE_HOOKUP_OBJECT (window, toolbarHandle, "toolbarHandle");
  GLADE_HOOKUP_OBJECT (window, toolbar, "toolbar");
  GLADE_HOOKUP_OBJECT (window, toolbarOpen, "toolbarOpen");
  GLADE_HOOKUP_OBJECT (window, toolbarReuse, "toolbarReuse");
  GLADE_HOOKUP_OBJECT (window, separatortoolitem1, "separatortoolitem1");
  GLADE_HOOKUP_OBJECT (window, toolbarSettings, "toolbarSettings");
  GLADE_HOOKUP_OBJECT (window, toolbarExit, "toolbarExit");
  GLADE_HOOKUP_OBJECT (window, windowPaned, "windowPaned");
  GLADE_HOOKUP_OBJECT (window, invisibleNotebook, "invisibleNotebook");
  GLADE_HOOKUP_OBJECT (window, pageOpen, "pageOpen");
  GLADE_HOOKUP_OBJECT (window, hbox27, "hbox27");
  GLADE_HOOKUP_OBJECT (window, aboutJigdoLogo, "aboutJigdoLogo");
  GLADE_HOOKUP_OBJECT (window, aboutBgnd, "aboutBgnd");
  GLADE_HOOKUP_OBJECT (window, vbox9, "vbox9");
  GLADE_HOOKUP_OBJECT (window, aboutJigdoLabel, "aboutJigdoLabel");
  GLADE_HOOKUP_OBJECT (window, aboutJigdoButton, "aboutJigdoButton");
  GLADE_HOOKUP_OBJECT (window, alignment11, "alignment11");
  GLADE_HOOKUP_OBJECT (window, aboutJigdoButtonLabel, "aboutJigdoButtonLabel");
  GLADE_HOOKUP_OBJECT (window, hbox24, "hbox24");
  GLADE_HOOKUP_OBJECT (window, vbox8, "vbox8");
  GLADE_HOOKUP_OBJECT (window, openButton, "openButton");
  GLADE_HOOKUP_OBJECT (window, alignment12, "alignment12");
  GLADE_HOOKUP_OBJECT (window, hbox95, "hbox95");
  GLADE_HOOKUP_OBJECT (window, image27, "image27");
  GLADE_HOOKUP_OBJECT (window, label247, "label247");
  GLADE_HOOKUP_OBJECT (window, table3, "table3");
  GLADE_HOOKUP_OBJECT (window, open_sourceURLLabel, "open_sourceURLLabel");
  GLADE_HOOKUP_OBJECT (window, open_destinationLabel, "open_destinationLabel");
  GLADE_HOOKUP_OBJECT (window, open_URL, "open_URL");
  GLADE_HOOKUP_OBJECT (window, open_dest, "open_dest");
  GLADE_HOOKUP_OBJECT (window, open_URLSel, "open_URLSel");
  GLADE_HOOKUP_OBJECT (window, open_destSel, "open_destSel");
  GLADE_HOOKUP_OBJECT (window, enterUrlLabel, "enterUrlLabel");
  GLADE_HOOKUP_OBJECT (window, pageOpenLabel, "pageOpenLabel");
  GLADE_HOOKUP_OBJECT (window, pageReuse, "pageReuse");
  GLADE_HOOKUP_OBJECT (window, scrolledwindow24, "scrolledwindow24");
  GLADE_HOOKUP_OBJECT (window, viewport6, "viewport6");
  GLADE_HOOKUP_OBJECT (window, label156, "label156");
  GLADE_HOOKUP_OBJECT (window, hbox53, "hbox53");
  GLADE_HOOKUP_OBJECT (window, reuse_path, "reuse_path");
  GLADE_HOOKUP_OBJECT (window, reuse_pathSel, "reuse_pathSel");
  GLADE_HOOKUP_OBJECT (window, hbox58, "hbox58");
  GLADE_HOOKUP_OBJECT (window, reuse_scanButton, "reuse_scanButton");
  GLADE_HOOKUP_OBJECT (window, hbox59, "hbox59");
  GLADE_HOOKUP_OBJECT (window, label161, "label161");
  GLADE_HOOKUP_OBJECT (window, reuse_clearButton, "reuse_clearButton");
  GLADE_HOOKUP_OBJECT (window, pageReuseLabel, "pageReuseLabel");
  GLADE_HOOKUP_OBJECT (window, pageDownload, "pageDownload");
  GLADE_HOOKUP_OBJECT (window, hbox96, "hbox96");
  GLADE_HOOKUP_OBJECT (window, _2lineshighlabelforsettingminimumheight, "_2lineshighlabelforsettingminimumheight");
  GLADE_HOOKUP_OBJECT (window, download_buttonInfo, "download_buttonInfo");
  GLADE_HOOKUP_OBJECT (window, vbox42, "vbox42");
  GLADE_HOOKUP_OBJECT (window, hbox83, "hbox83");
  GLADE_HOOKUP_OBJECT (window, download_closeButton, "download_closeButton");
  GLADE_HOOKUP_OBJECT (window, image14, "image14");
  GLADE_HOOKUP_OBJECT (window, download_restartButton, "download_restartButton");
  GLADE_HOOKUP_OBJECT (window, image15, "image15");
  GLADE_HOOKUP_OBJECT (window, label211, "label211");
  GLADE_HOOKUP_OBJECT (window, download_stopButton, "download_stopButton");
  GLADE_HOOKUP_OBJECT (window, image10, "image10");
  GLADE_HOOKUP_OBJECT (window, download_pauseButton, "download_pauseButton");
  GLADE_HOOKUP_OBJECT (window, image11, "image11");
  GLADE_HOOKUP_OBJECT (window, download_startButton, "download_startButton");
  GLADE_HOOKUP_OBJECT (window, image13, "image13");
  GLADE_HOOKUP_OBJECT (window, scrolledwindow23, "scrolledwindow23");
  GLADE_HOOKUP_OBJECT (window, viewport5, "viewport5");
  GLADE_HOOKUP_OBJECT (window, table5, "table5");
  GLADE_HOOKUP_OBJECT (window, download_dest, "download_dest");
  GLADE_HOOKUP_OBJECT (window, download_progress, "download_progress");
  GLADE_HOOKUP_OBJECT (window, label151, "label151");
  GLADE_HOOKUP_OBJECT (window, download_status, "download_status");
  GLADE_HOOKUP_OBJECT (window, download_URL, "download_URL");
  GLADE_HOOKUP_OBJECT (window, label131, "label131");
  GLADE_HOOKUP_OBJECT (window, label133, "label133");
  GLADE_HOOKUP_OBJECT (window, label149, "label149");
  GLADE_HOOKUP_OBJECT (window, pageDownloadLabel, "pageDownloadLabel");
  GLADE_HOOKUP_OBJECT (window, pageJigdo, "pageJigdo");
  GLADE_HOOKUP_OBJECT (window, vbox34, "vbox34");
  GLADE_HOOKUP_OBJECT (window, scrolledwindow20, "scrolledwindow20");
  GLADE_HOOKUP_OBJECT (window, viewport2, "viewport2");
  GLADE_HOOKUP_OBJECT (window, jigdo_InfoVbox, "jigdo_InfoVbox");
  GLADE_HOOKUP_OBJECT (window, label222, "label222");
  GLADE_HOOKUP_OBJECT (window, hseparator5, "hseparator5");
  GLADE_HOOKUP_OBJECT (window, jigdo_ShortInfo, "jigdo_ShortInfo");
  GLADE_HOOKUP_OBJECT (window, jigdo_Info, "jigdo_Info");
  GLADE_HOOKUP_OBJECT (window, hbuttonbox9, "hbuttonbox9");
  GLADE_HOOKUP_OBJECT (window, button119, "button119");
  GLADE_HOOKUP_OBJECT (window, alignment3, "alignment3");
  GLADE_HOOKUP_OBJECT (window, hbox85, "hbox85");
  GLADE_HOOKUP_OBJECT (window, image16, "image16");
  GLADE_HOOKUP_OBJECT (window, label223, "label223");
  GLADE_HOOKUP_OBJECT (window, label213, "label213");
  GLADE_HOOKUP_OBJECT (window, vbox36, "vbox36");
  GLADE_HOOKUP_OBJECT (window, vbox39, "vbox39");
  GLADE_HOOKUP_OBJECT (window, label229, "label229");
  GLADE_HOOKUP_OBJECT (window, label230, "label230");
  GLADE_HOOKUP_OBJECT (window, hbox88, "hbox88");
  GLADE_HOOKUP_OBJECT (window, entry23, "entry23");
  GLADE_HOOKUP_OBJECT (window, alignment6, "alignment6");
  GLADE_HOOKUP_OBJECT (window, button127, "button127");
  GLADE_HOOKUP_OBJECT (window, button128, "button128");
  GLADE_HOOKUP_OBJECT (window, alignment5, "alignment5");
  GLADE_HOOKUP_OBJECT (window, hbox89, "hbox89");
  GLADE_HOOKUP_OBJECT (window, image23, "image23");
  GLADE_HOOKUP_OBJECT (window, label227, "label227");
  GLADE_HOOKUP_OBJECT (window, hbuttonbox10, "hbuttonbox10");
  GLADE_HOOKUP_OBJECT (window, button120, "button120");
  GLADE_HOOKUP_OBJECT (window, button121, "button121");
  GLADE_HOOKUP_OBJECT (window, alignment4, "alignment4");
  GLADE_HOOKUP_OBJECT (window, hbox86, "hbox86");
  GLADE_HOOKUP_OBJECT (window, image17, "image17");
  GLADE_HOOKUP_OBJECT (window, label224, "label224");
  GLADE_HOOKUP_OBJECT (window, label217, "label217");
  GLADE_HOOKUP_OBJECT (window, vbox40, "vbox40");
  GLADE_HOOKUP_OBJECT (window, vbox41, "vbox41");
  GLADE_HOOKUP_OBJECT (window, label239, "label239");
  GLADE_HOOKUP_OBJECT (window, radiobutton1, "radiobutton1");
  GLADE_HOOKUP_OBJECT (window, hbox93, "hbox93");
  GLADE_HOOKUP_OBJECT (window, radiobutton2, "radiobutton2");
  GLADE_HOOKUP_OBJECT (window, combo1, "combo1");
  GLADE_HOOKUP_OBJECT (window, combo_entry1, "combo_entry1");
  GLADE_HOOKUP_OBJECT (window, hbox94, "hbox94");
  GLADE_HOOKUP_OBJECT (window, alignment10, "alignment10");
  GLADE_HOOKUP_OBJECT (window, radiobutton3, "radiobutton3");
  GLADE_HOOKUP_OBJECT (window, scrolledwindow21, "scrolledwindow21");
  GLADE_HOOKUP_OBJECT (window, viewport3, "viewport3");
  GLADE_HOOKUP_OBJECT (window, table8, "table8");
  GLADE_HOOKUP_OBJECT (window, label243, "label243");
  GLADE_HOOKUP_OBJECT (window, label244, "label244");
  GLADE_HOOKUP_OBJECT (window, label245, "label245");
  GLADE_HOOKUP_OBJECT (window, label246, "label246");
  GLADE_HOOKUP_OBJECT (window, combo2, "combo2");
  GLADE_HOOKUP_OBJECT (window, combo_entry2, "combo_entry2");
  GLADE_HOOKUP_OBJECT (window, combo3, "combo3");
  GLADE_HOOKUP_OBJECT (window, combo_entry3, "combo_entry3");
  GLADE_HOOKUP_OBJECT (window, hbuttonbox11, "hbuttonbox11");
  GLADE_HOOKUP_OBJECT (window, button131, "button131");
  GLADE_HOOKUP_OBJECT (window, button132, "button132");
  GLADE_HOOKUP_OBJECT (window, alignment9, "alignment9");
  GLADE_HOOKUP_OBJECT (window, hbox92, "hbox92");
  GLADE_HOOKUP_OBJECT (window, image25, "image25");
  GLADE_HOOKUP_OBJECT (window, label242, "label242");
  GLADE_HOOKUP_OBJECT (window, label218, "label218");
  GLADE_HOOKUP_OBJECT (window, vbox37, "vbox37");
  GLADE_HOOKUP_OBJECT (window, hbox97, "hbox97");
  GLADE_HOOKUP_OBJECT (window, _2lineshighforsettingminheight, "_2lineshighforsettingminheight");
  GLADE_HOOKUP_OBJECT (window, jigdo_buttonInfo, "jigdo_buttonInfo");
  GLADE_HOOKUP_OBJECT (window, vbox43, "vbox43");
  GLADE_HOOKUP_OBJECT (window, hbox98, "hbox98");
  GLADE_HOOKUP_OBJECT (window, jigdo_closeButton, "jigdo_closeButton");
  GLADE_HOOKUP_OBJECT (window, image28, "image28");
  GLADE_HOOKUP_OBJECT (window, jigdo_restartButton, "jigdo_restartButton");
  GLADE_HOOKUP_OBJECT (window, image29, "image29");
  GLADE_HOOKUP_OBJECT (window, label251, "label251");
  GLADE_HOOKUP_OBJECT (window, jigdo_stopButton, "jigdo_stopButton");
  GLADE_HOOKUP_OBJECT (window, image30, "image30");
  GLADE_HOOKUP_OBJECT (window, jigdo_pauseButton, "jigdo_pauseButton");
  GLADE_HOOKUP_OBJECT (window, image31, "image31");
  GLADE_HOOKUP_OBJECT (window, jigdo_startButton, "jigdo_startButton");
  GLADE_HOOKUP_OBJECT (window, image32, "image32");
  GLADE_HOOKUP_OBJECT (window, scrolledwindow22, "scrolledwindow22");
  GLADE_HOOKUP_OBJECT (window, viewport4, "viewport4");
  GLADE_HOOKUP_OBJECT (window, table7, "table7");
  GLADE_HOOKUP_OBJECT (window, label233, "label233");
  GLADE_HOOKUP_OBJECT (window, label235, "label235");
  GLADE_HOOKUP_OBJECT (window, label237, "label237");
  GLADE_HOOKUP_OBJECT (window, label231, "label231");
  GLADE_HOOKUP_OBJECT (window, jigdo_status, "jigdo_status");
  GLADE_HOOKUP_OBJECT (window, jigdo_progress, "jigdo_progress");
  GLADE_HOOKUP_OBJECT (window, jigdo_dest, "jigdo_dest");
  GLADE_HOOKUP_OBJECT (window, jigdo_URL, "jigdo_URL");
  GLADE_HOOKUP_OBJECT (window, label219, "label219");
  GLADE_HOOKUP_OBJECT (window, pageJigdoLabel, "pageJigdoLabel");
  GLADE_HOOKUP_OBJECT (window, progressScroll, "progressScroll");
  GLADE_HOOKUP_OBJECT (window, jobs, "jobs");
  GLADE_HOOKUP_OBJECT_NO_REF (window, tooltips, "tooltips");

  gtk_widget_grab_focus (openButton);
  gtk_widget_grab_default (openButton);
  return window;
}

GtkWidget*
GUI::Filesel::create()
{

  filesel = gtk_file_selection_new (_("Select File"));
  gtk_container_set_border_width (GTK_CONTAINER (filesel), 10);
  gtk_window_set_type_hint (GTK_WINDOW (filesel), GDK_WINDOW_TYPE_HINT_DIALOG);

  okButton = GTK_FILE_SELECTION (filesel)->ok_button;
  gtk_widget_show (okButton);
  GTK_WIDGET_SET_FLAGS (okButton, GTK_CAN_DEFAULT);

  cancelButton = GTK_FILE_SELECTION (filesel)->cancel_button;
  gtk_widget_show (cancelButton);
  GTK_WIDGET_SET_FLAGS (cancelButton, GTK_CAN_DEFAULT);

  g_signal_connect_swapped ((gpointer) filesel, "delete_event",
                            G_CALLBACK (gtk_widget_hide),
                            GTK_OBJECT (filesel));
  g_signal_connect_swapped ((gpointer) cancelButton, "clicked",
                            G_CALLBACK (gtk_widget_hide),
                            GTK_OBJECT (filesel));

  /* Store pointers to all widgets, for use by lookup_widget(). */
  GLADE_HOOKUP_OBJECT_NO_REF (filesel, filesel, "filesel");
  GLADE_HOOKUP_OBJECT_NO_REF (filesel, okButton, "okButton");
  GLADE_HOOKUP_OBJECT_NO_REF (filesel, cancelButton, "cancelButton");

  return filesel;
}

GtkWidget*
GUI::Settings::create()
{
  GtkWidget *vbox26;
  GtkWidget *hbox40;
  GtkWidget *hbuttonbox7;
  GtkWidget *button75;
  GtkWidget *button76;
  GtkWidget *button79;
  GtkWidget *vbox28;
  GtkWidget *label119;
  GtkWidget *hbox41;
  GtkWidget *label120;
  GtkWidget *spinbutton2;
  GtkWidget *checkbutton9;
  GtkWidget *checkbutton10;
  GtkWidget *hbox42;
  GtkWidget *checkbutton11;
  GtkWidget *checkbutton12;
  GtkWidget *hbox77;
  GtkWidget *label206;
  GtkWidget *spinbutton3;
  GtkWidget *table4;
  GtkWidget *label121;
  GtkWidget *label122;
  GtkWidget *label123;
  GtkWidget *entry15;
  GtkWidget *entry16;
  GtkWidget *label193;
  GtkWidget *entry17;
  GtkWidget *entry22;
  GtkWidget *hseparator4;

  tooltips = gtk_tooltips_new ();

  settings = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title (GTK_WINDOW (settings), _("Settings - Jigsaw Download"));

  vbox26 = gtk_vbox_new (FALSE, 0);
  gtk_widget_show (vbox26);
  gtk_container_add (GTK_CONTAINER (settings), vbox26);

  hbox40 = gtk_hbox_new (TRUE, 5);
  gtk_widget_show (hbox40);
  gtk_box_pack_end (GTK_BOX (vbox26), hbox40, FALSE, TRUE, 0);
  gtk_container_set_border_width (GTK_CONTAINER (hbox40), 10);

  hbuttonbox7 = gtk_hbutton_box_new ();
  gtk_widget_show (hbuttonbox7);
  gtk_box_pack_start (GTK_BOX (hbox40), hbuttonbox7, TRUE, TRUE, 0);
  gtk_button_box_set_layout (GTK_BUTTON_BOX (hbuttonbox7), GTK_BUTTONBOX_END);
  gtk_box_set_spacing (GTK_BOX (hbuttonbox7), 8);

  button75 = gtk_button_new_with_mnemonic (_("OK"));
  gtk_widget_show (button75);
  gtk_container_add (GTK_CONTAINER (hbuttonbox7), button75);
  GTK_WIDGET_SET_FLAGS (button75, GTK_CAN_DEFAULT);

  button76 = gtk_button_new_with_mnemonic (_("Apply"));
  gtk_widget_show (button76);
  gtk_container_add (GTK_CONTAINER (hbuttonbox7), button76);
  GTK_WIDGET_SET_FLAGS (button76, GTK_CAN_DEFAULT);

  button79 = gtk_button_new_with_mnemonic (_("Cancel"));
  gtk_widget_show (button79);
  gtk_container_add (GTK_CONTAINER (hbuttonbox7), button79);
  GTK_WIDGET_SET_FLAGS (button79, GTK_CAN_DEFAULT);

  vbox28 = gtk_vbox_new (FALSE, 0);
  gtk_widget_show (vbox28);
  gtk_box_pack_start (GTK_BOX (vbox26), vbox28, TRUE, TRUE, 0);
  gtk_container_set_border_width (GTK_CONTAINER (vbox28), 6);

  label119 = gtk_label_new (_("Settings"));
  gtk_widget_show (label119);
  gtk_box_pack_start (GTK_BOX (vbox28), label119, TRUE, FALSE, 0);
  gtk_label_set_justify (GTK_LABEL (label119), GTK_JUSTIFY_CENTER);

  hbox41 = gtk_hbox_new (FALSE, 4);
  gtk_widget_show (hbox41);
  gtk_box_pack_start (GTK_BOX (vbox28), hbox41, TRUE, TRUE, 0);

  label120 = gtk_label_new (_("Maximum number of simultaneous downloads"));
  gtk_widget_show (label120);
  gtk_box_pack_start (GTK_BOX (hbox41), label120, FALSE, FALSE, 0);
  gtk_label_set_justify (GTK_LABEL (label120), GTK_JUSTIFY_RIGHT);
  gtk_misc_set_padding (GTK_MISC (label120), 2, 0);

  spinbutton2_adj = gtk_adjustment_new (1, 1, 5, 1, 10, 10);
  spinbutton2 = gtk_spin_button_new (GTK_ADJUSTMENT (spinbutton2_adj), 1, 0);
  gtk_widget_show (spinbutton2);
  gtk_box_pack_start (GTK_BOX (hbox41), spinbutton2, FALSE, FALSE, 0);

  checkbutton9 = gtk_check_button_new_with_mnemonic (_("Display tooltips"));
  gtk_widget_show (checkbutton9);
  gtk_box_pack_start (GTK_BOX (vbox28), checkbutton9, FALSE, FALSE, 0);
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (checkbutton9), TRUE);

  checkbutton10 = gtk_check_button_new_with_mnemonic (_("Switch to download screen once download of a .jigdo file is complete"));
  gtk_widget_show (checkbutton10);
  gtk_box_pack_start (GTK_BOX (vbox28), checkbutton10, FALSE, FALSE, 0);
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (checkbutton10), TRUE);

  hbox42 = gtk_hbox_new (FALSE, 0);
  gtk_widget_show (hbox42);
  gtk_box_pack_start (GTK_BOX (vbox28), hbox42, FALSE, FALSE, 0);

  checkbutton11 = gtk_check_button_new_with_mnemonic (_("Save .jigdo files in destination directory"));
  gtk_widget_show (checkbutton11);
  gtk_box_pack_start (GTK_BOX (hbox42), checkbutton11, TRUE, FALSE, 0);

  checkbutton12 = gtk_check_button_new_with_mnemonic (_("Save .template files in destination directory"));
  gtk_widget_show (checkbutton12);
  gtk_box_pack_start (GTK_BOX (hbox42), checkbutton12, TRUE, FALSE, 0);

  hbox77 = gtk_hbox_new (FALSE, 4);
  gtk_widget_show (hbox77);
  gtk_box_pack_start (GTK_BOX (vbox28), hbox77, TRUE, TRUE, 0);

  label206 = gtk_label_new (_("Remove finished download jobs from list after (seconds)"));
  gtk_widget_show (label206);
  gtk_box_pack_start (GTK_BOX (hbox77), label206, FALSE, FALSE, 0);
  gtk_label_set_justify (GTK_LABEL (label206), GTK_JUSTIFY_RIGHT);
  gtk_misc_set_padding (GTK_MISC (label206), 2, 0);

  spinbutton3_adj = gtk_adjustment_new (1, 1, 100000000, 1, 10, 10);
  spinbutton3 = gtk_spin_button_new (GTK_ADJUSTMENT (spinbutton3_adj), 1, 0);
  gtk_widget_show (spinbutton3);
  gtk_box_pack_start (GTK_BOX (hbox77), spinbutton3, FALSE, FALSE, 0);

  table4 = gtk_table_new (4, 2, FALSE);
  gtk_widget_show (table4);
  gtk_box_pack_start (GTK_BOX (vbox28), table4, TRUE, TRUE, 0);
  gtk_table_set_row_spacings (GTK_TABLE (table4), 2);
  gtk_table_set_col_spacings (GTK_TABLE (table4), 4);

  label121 = gtk_label_new (_("HTTP proxy"));
  gtk_widget_show (label121);
  gtk_table_attach (GTK_TABLE (table4), label121, 0, 1, 0, 1,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);
  gtk_label_set_justify (GTK_LABEL (label121), GTK_JUSTIFY_RIGHT);
  gtk_misc_set_alignment (GTK_MISC (label121), 7.45058e-09, 0.5);
  gtk_misc_set_padding (GTK_MISC (label121), 2, 0);

  label122 = gtk_label_new (_("FTP proxy"));
  gtk_widget_show (label122);
  gtk_table_attach (GTK_TABLE (table4), label122, 0, 1, 1, 2,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);
  gtk_label_set_justify (GTK_LABEL (label122), GTK_JUSTIFY_CENTER);
  gtk_misc_set_alignment (GTK_MISC (label122), 7.45058e-09, 0.5);
  gtk_misc_set_padding (GTK_MISC (label122), 2, 0);

  label123 = gtk_label_new (_("no proxy"));
  gtk_widget_show (label123);
  gtk_table_attach (GTK_TABLE (table4), label123, 0, 1, 2, 3,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);
  gtk_misc_set_alignment (GTK_MISC (label123), 7.45058e-09, 0.5);
  gtk_misc_set_padding (GTK_MISC (label123), 2, 0);

  entry15 = gtk_entry_new ();
  gtk_widget_show (entry15);
  gtk_table_attach (GTK_TABLE (table4), entry15, 1, 2, 0, 1,
                    (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);

  entry16 = gtk_entry_new ();
  gtk_widget_show (entry16);
  gtk_table_attach (GTK_TABLE (table4), entry16, 1, 2, 1, 2,
                    (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);

  label193 = gtk_label_new (_("Default servers"));
  gtk_widget_show (label193);
  gtk_table_attach (GTK_TABLE (table4), label193, 0, 1, 3, 4,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);
  gtk_label_set_justify (GTK_LABEL (label193), GTK_JUSTIFY_CENTER);
  gtk_misc_set_alignment (GTK_MISC (label193), 0, 0.5);

  entry17 = gtk_entry_new ();
  gtk_widget_show (entry17);
  gtk_table_attach (GTK_TABLE (table4), entry17, 1, 2, 2, 3,
                    (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);
  gtk_tooltips_set_tip (tooltips, entry17, _("Enter a list of domains (separated with spaces) for which you do not want to use the proxies you set up above. Example: \"mycompany.com local.network.lan\""), NULL);

  entry22 = gtk_entry_new ();
  gtk_widget_show (entry22);
  gtk_table_attach (GTK_TABLE (table4), entry22, 1, 2, 3, 4,
                    (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);
  gtk_tooltips_set_tip (tooltips, entry22, _("Enter a list of domains (separated with spaces) for which you do not want to use the proxies you set up above. Example: \"mycompany.com local.network.lan\""), NULL);

  hseparator4 = gtk_hseparator_new ();
  gtk_widget_show (hseparator4);
  gtk_box_pack_start (GTK_BOX (vbox26), hseparator4, FALSE, FALSE, 0);

  /* Store pointers to all widgets, for use by lookup_widget(). */
  GLADE_HOOKUP_OBJECT_NO_REF (settings, settings, "settings");
  GLADE_HOOKUP_OBJECT (settings, vbox26, "vbox26");
  GLADE_HOOKUP_OBJECT (settings, hbox40, "hbox40");
  GLADE_HOOKUP_OBJECT (settings, hbuttonbox7, "hbuttonbox7");
  GLADE_HOOKUP_OBJECT (settings, button75, "button75");
  GLADE_HOOKUP_OBJECT (settings, button76, "button76");
  GLADE_HOOKUP_OBJECT (settings, button79, "button79");
  GLADE_HOOKUP_OBJECT (settings, vbox28, "vbox28");
  GLADE_HOOKUP_OBJECT (settings, label119, "label119");
  GLADE_HOOKUP_OBJECT (settings, hbox41, "hbox41");
  GLADE_HOOKUP_OBJECT (settings, label120, "label120");
  GLADE_HOOKUP_OBJECT (settings, spinbutton2, "spinbutton2");
  GLADE_HOOKUP_OBJECT (settings, checkbutton9, "checkbutton9");
  GLADE_HOOKUP_OBJECT (settings, checkbutton10, "checkbutton10");
  GLADE_HOOKUP_OBJECT (settings, hbox42, "hbox42");
  GLADE_HOOKUP_OBJECT (settings, checkbutton11, "checkbutton11");
  GLADE_HOOKUP_OBJECT (settings, checkbutton12, "checkbutton12");
  GLADE_HOOKUP_OBJECT (settings, hbox77, "hbox77");
  GLADE_HOOKUP_OBJECT (settings, label206, "label206");
  GLADE_HOOKUP_OBJECT (settings, spinbutton3, "spinbutton3");
  GLADE_HOOKUP_OBJECT (settings, table4, "table4");
  GLADE_HOOKUP_OBJECT (settings, label121, "label121");
  GLADE_HOOKUP_OBJECT (settings, label122, "label122");
  GLADE_HOOKUP_OBJECT (settings, label123, "label123");
  GLADE_HOOKUP_OBJECT (settings, entry15, "entry15");
  GLADE_HOOKUP_OBJECT (settings, entry16, "entry16");
  GLADE_HOOKUP_OBJECT (settings, label193, "label193");
  GLADE_HOOKUP_OBJECT (settings, entry17, "entry17");
  GLADE_HOOKUP_OBJECT (settings, entry22, "entry22");
  GLADE_HOOKUP_OBJECT (settings, hseparator4, "hseparator4");
  GLADE_HOOKUP_OBJECT_NO_REF (settings, tooltips, "tooltips");

  gtk_widget_grab_default (button75);
  return settings;
}

GtkWidget*
GUI::License::create()
{

  license = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title (GTK_WINDOW (license), _("License for Jigsaw Download"));
  gtk_window_set_default_size (GTK_WINDOW (license), 516, 300);

  licenseScroll = gtk_scrolled_window_new (NULL, NULL);
  gtk_widget_show (licenseScroll);
  gtk_container_add (GTK_CONTAINER (license), licenseScroll);
  gtk_container_set_border_width (GTK_CONTAINER (licenseScroll), 4);
  GTK_WIDGET_UNSET_FLAGS (licenseScroll, GTK_CAN_FOCUS);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (licenseScroll), GTK_POLICY_NEVER, GTK_POLICY_ALWAYS);
  gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (licenseScroll), GTK_SHADOW_IN);

  licenseText = gtk_text_view_new ();
  gtk_widget_show (licenseText);
  gtk_container_add (GTK_CONTAINER (licenseScroll), licenseText);
  gtk_text_view_set_editable (GTK_TEXT_VIEW (licenseText), FALSE);
  gtk_text_view_set_wrap_mode (GTK_TEXT_VIEW (licenseText), GTK_WRAP_WORD);
  gtk_text_view_set_cursor_visible (GTK_TEXT_VIEW (licenseText), FALSE);
  gtk_text_view_set_left_margin (GTK_TEXT_VIEW (licenseText), 4);
  gtk_text_view_set_right_margin (GTK_TEXT_VIEW (licenseText), 4);

  g_signal_connect_swapped ((gpointer) license, "delete_event",
                            G_CALLBACK (gtk_widget_hide),
                            GTK_OBJECT (license));

  /* Store pointers to all widgets, for use by lookup_widget(). */
  GLADE_HOOKUP_OBJECT_NO_REF (license, license, "license");
  GLADE_HOOKUP_OBJECT (license, licenseScroll, "licenseScroll");
  GLADE_HOOKUP_OBJECT (license, licenseText, "licenseText");

  return license;
}

