/* $Id: makeimage.cc,v 1.1.1.1 2003/07/04 22:29:52 atterer Exp $ -*- C++ -*-
  __   _
  |_) /|  Copyright (C) 2003  |  richard@
  | \/¯|  Richard Atterer     |  atterer.org
  ¯ '` ¯
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License, version 2. See
  the file COPYING for details.

  Download & interpret .jigdo, download parts, assemble image

*/

#include <config.h>

#include <glib.h>
#include <iostream>
#include <string.h>

#include <makeimage.hh>
//______________________________________________________________________
